import numpy as np
from .random_state import random_state


def random_lattice(n: int, seed=None):
    '''
    Description:
    
        http://ka.karlin.mff.cuni.cz/jezek/093/random.pdf
    
    This algorihtm works after fixing three issues:
    
        1. k might be zero when rnd(k) is called
           (fixed with k = max(k,1))
        2. u might be undefined at range(u)
           (fixed with u = 1)
        3. The output guarantees unique lub for each pair
           but not unique bottom element
           (fixed in Poset class)
    '''
    R = random_state(seed=seed)
    J = np.zeros((n, n), dtype=int) - 1
    for i in range(n):
        J[i, i] = i

    M = np.zeros((n,), dtype=np.int64)
    Q = np.zeros((n,), dtype=np.int64)

    rnd = lambda k: R.randint(0, k)

    def S(i: int):
        return max(2, int(np.ceil(np.sqrt(i))))

    def FindMax(i: int):
        k = 0
        for j in range(i):
            s = 1
            for a in range(i):
                if a != j and J[a, j] == a:
                    s = 0
            if s != 0:
                M[k] = j
                k += 1
        k = max(k, 1)  # This line did not exist
        a = rnd(k)
        a += 1
        for j in range(k):
            Q[j] = 0
        for s in range(a):
            j = rnd(k)
            if Q[j]:
                s -= 1
            else:
                Q[j] = 1
        return k

    def Work(i):
        if i == n - 1:
            for j in range(n):
                for l in range(n):
                    if J[j, l] == -1:
                        J[j, l] = n - 1
            return
        q = S(n - i)
        if i == 1:
            u = 1
            M[0] = 0
            Q[0] = 1
        elif rnd(q) == 0:
            u = FindMax(i)
        else:  # This case did not exist
            u = 1

        for j in range(u):
            if Q[j]:
                J[M[j], i] = i
                J[i, M[j]] = i
        w = 1
        while w:
            w = 0
            for j in range(i):
                if (J[j, i] == i):
                    for s in range(i):
                        if J[s, j] == j and J[s, i] != i:
                            w = 1
                            J[s, i] = i
                            J[i, s] = i
            for j in range(i):
                if (J[j, i] == i):
                    for l in range(i):
                        if J[l, i] == i:
                            s = J[j, l]
                            if s != -1 and J[s, i] != i:
                                w = 1
                                J[s, i] = i
                                J[i, s] = i
        for j in range(i):
            if J[j, i] == i:
                for l in range(i):
                    if J[l, i] == i and J[j, l] == -1:
                        J[j, l] = i
                        J[l, j] = i
        return

    for i in range(n):
        Work(i)

    return J


from __future__ import annotations
from fractions import Fraction
import itertools
from typing import TYPE_CHECKING, Iterable, Iterator, List, Optional, Sequence, Tuple, TypeVar, Union, cast

from avispa_lattices import base
if TYPE_CHECKING:
    from ..base import Lattice, Poset, Relation
    T_Relation = TypeVar('T_Relation', Relation, Poset, Lattice)

from math import factorial
from collections import deque
import numpy as np
from ..utils.numpy_types import npUInt64Matrix, npInt64Array
import pyhash
from .graph import heights

Ints = Union[Sequence[int], npInt64Array]

_hasher = pyhash.xx_64(seed=0)


def hasher(ints: Sequence[int]):
    '''
    Fast numeric hashing function that is consistent across runs.
    Independent of PYTHONHASHSEED unlike Python's hash.
    The output space is range(2**63), i.e. 1e18 approximately 
    '''
    uint64hash = _hasher(str(ints)[1:-1])
    int64hash = uint64hash >> 1  # Prevent overflow
    return int64hash


def _hash(P: Poset, rounds: int):
    elems = _hash_elems(P, rounds=rounds, salt=0)
    return hasher(sorted(elems))


def _hash_elems(P: Poset, rounds: int, salt: int):
    mat: npUInt64Matrix = P.leq.astype(np.int64)
    with np.errstate(over='ignore'):
        H = hash_perm_invariant(P, salt + mat)
        for repeat in range(rounds):
            mat += np.matmul(H[:, None], H[None, :])
            H = hash_perm_invariant(P, salt + mat)
    return cast(npInt64Array, H)


def hash_perm_invariant(P: Poset, mat: npUInt64Matrix):
    h = lambda l: hasher(sorted(l))
    a = [hasher((h(mat[:, i]), h(mat[i, :]))) for i in range(P.n)]
    return np.array(a, dtype=int)


def find_isomorphism(P: Poset, other: Poset):
    # Quick check:
    if P.n != other.n or hash(P) != hash(other):
        return None

    # Find the isomorphism
    n = P.n
    A = P.leq
    B = other.leq
    IJ = [(i, j) for i in range(n) for j in range(n)]

    def is_isomorphism(f):
        return all(A[i, j] == B[f[i], f[j]] for i, j in IJ)

    Ah = P.hash_elems
    Bh = other.hash_elems
    total, it = isomorphism_candidates(Ah, Bh)
    if total > n**2:
        # Try to hash more deeply to separate
        AAh = Ah + _hash_elems(P, rounds=4, salt=1)
        BBh = Bh + _hash_elems(other, rounds=4, salt=1)
        total, it = isomorphism_candidates(AAh, BBh)

    f: List[int]
    for f in it:
        if is_isomorphism(f):
            return f
    return None


def reindex(P: T_Relation, f, inverse=False, reset_labels=False) -> T_Relation:
    'Reindexed copy of P such that i is to P as f[i] to out'
    'If inverse==True, then f[i] is to P as i to out'
    n = P.n
    assert len(f) == n and sorted(set(f)) == list(
        range(n)), f'Invalid permutation {f}'
    if inverse:
        inv = [0] * n
        for i in range(n):
            inv[f[i]] = i
        f = inv
    leq = P.leq
    out = np.zeros_like(leq)
    for i in range(n):
        for j in range(n):
            out[f[i], f[j]] = leq[i, j]
    out.flags.writeable = False
    out_labels: Optional[Sequence[str]]
    if reset_labels:
        out_labels = None
    else:
        P_labels = P._labels
        out_labels = ['' for i in range(n)]
        for i in range(n):
            out_labels[f[i]] = P_labels[i]
        out_labels = tuple(out_labels)
    return P.__class__(out, False, labels=out_labels)


def canonical_rank_OLD(P: Poset):
    'equivalent poset with enumerated labels and stable order'
    n = P.n
    group_by = {h: [] for h in range(n)}
    h = heights(P)
    for i in range(n):
        group_by[h[i]].append(i)
    pa = P.parents
    ch = P.children
    nleq = P.leq.sum(axis=0)
    ngeq = P.leq.sum(axis=1)
    order = list(zip(nleq, -ngeq, P.hash_elems, P._labels, range(n)))

    def key(i):
        t = tuple(sorted((rank[i] for i in ch[i])))
        return (t, len(pa[i]), order[i])

    topo = []
    rank = [-1] * n
    for h in range(n):
        for i in sorted(group_by[h], key=key):
            rank[i] = len(topo)
            topo.append(i)
    return rank


def canonical_rank(P: Poset):
    'equivalent poset with enumerated labels and stable order'
    n = P.n
    pa = P.parents
    ch = P.children
    nleq = P.leq.sum(axis=0)
    ngeq = P.leq.sum(axis=1)
    nchild = P.child.sum(axis=0)
    vis = set()

    _base_score = [Fraction(nleq[i] + 1, ngeq[i] + 1) for i in range(n)]
    base_score = list(zip(_base_score, -nchild, P.hash_elems))

    last = 0
    rank = [-1] * n
    pa_rank = [n] * n
    vis = set()
    next_layer = set()
    for i in range(n):
        if len(ch[i]) == 0:
            if i not in vis:
                next_layer.add(i)
                vis.add(i)

    while next_layer:
        layer = next_layer
        # Compute ranks
        for i in sorted(layer, key=lambda i: (pa_rank[i], base_score[i])):
            rank[i] = last
            last += 1
        # Compute next_layer
        next_layer = set()
        for i in layer:
            for j in pa[i]:
                if j not in vis:
                    next_layer.add(j)
                    vis.add(j)
                pa_rank[j] = min(pa_rank[j], rank[i])
    return rank


def isomorphism_candidates(hashesA: Ints, hashesB: Ints,
                           out=None) -> Tuple[int, Iterable[List[int]]]:
    '''
    Total number and iterator of all injective and surjective mappings
        f from range(n) to range(n)
    such that
        all(hashesA[i] == hashesB[f[i]] for i in range(n))
    where n = len(hashesA) = len(hashesB).
    '''
    n = len(hashesA)
    if len(hashesA) != len(hashesB):
        return 0, iter([])
    if out is not None:
        assert len(out) == n, f'Incompatible output shape'

    out = [None] * n if out is None else out
    out = cast(List[int], out)

    if sorted(hashesA) != sorted(hashesB):
        return 0, iter([])
    total = 1

    empty = lambda: cast(List[int], [])
    groups = {v: (empty(), empty()) for v in [*hashesA, *hashesB]}
    for i, v in enumerate(hashesA):
        groups[v][0].append(i)
    for i, v in enumerate(hashesB):
        groups[v][1].append(i)

    groups = [*groups.values()]
    for idxA, idxB in groups:
        if len(idxA) != len(idxB):
            return 0, iter([])
        total *= factorial(len(idxA))

    m = len(groups)

    def backtrack(group_i):
        if group_i == m:
            yield out
        else:
            gA, gB = groups[group_i]
            for gBperm in itertools.permutations(gB):
                for i, j in zip(gA, gBperm):
                    out[i] = j
                yield from backtrack(group_i + 1)

    return total, backtrack(0)



from __future__ import annotations

from typing import TYPE_CHECKING, Iterable, List, Optional, Tuple
import numpy as np

from ..utils import algorithm_random_poset_czech
from ..utils.random_state import random_state

from ..base import Lattice, Poset, Relation, Endomorphism


def random_poset(n: int, p: float, seed=None):
    '''
    Generates a random poset.
    All posets (modulo labels) have positive probability of being generated.
    If p is close to 0, the poset is very sparse.
    If p is close to 1, the poset is very dense.
    '''
    R = random_state(seed)
    rel = np.zeros((n, n), dtype=bool)
    for i in range(n):
        for j in range(i + 1, n):
            if R.random() < p:
                rel[i, j] = 1
    for i in range(n):
        rel[i, i] = 1
    rel.flags.writeable = False
    leq = Relation(rel, check=False).transitive_closure().leq
    poset = Poset(leq, False)
    return poset


def random_lattice_czech(n: int, seed=None):
    '''
    Description: http://ka.karlin.mff.cuni.cz/jezek/093/random.pdf
    '''
    lub = algorithm_random_poset_czech.random_lattice(n, seed)
    rel = (lub <= np.arange(n)[None, :])
    rel.flags.writeable = False
    L = Lattice(rel, False)
    if len(L.bottoms) >= 2:  # Collapse bottoms
        bottoms = L.bottoms
        bot = bottoms[0]
        rel.flags.writeable = True
        for i in bottoms:
            rel[bot, i] = True
        rel.flags.writeable = False
        L = Lattice(rel, check=False).transitive_closure()
        #poset.assert_lattice()
    return L


def random_f_monotone_A(L: Lattice, seed=None,
                        _max_arbitrary_fi: Optional[int] = None):
    R = random_state(seed)
    f: Endomorphism = [-1] * L.n
    n_above = L.leq.sum(axis=1)
    p_above = n_above / n_above.sum()
    elems = np.arange(L.n)
    f[:] = R.choice(elems, L.n, p=p_above)
    if _max_arbitrary_fi is not None and _max_arbitrary_fi < L.n:
        for i in R.choice(elems, L.n - _max_arbitrary_fi):
            f[i] = L.bottom
    f[L.bottom] = L.bottom
    for i in L.toposort_bottom_up:
        min_fi = L.lub_of_many(*[f[k] for k in L.children[i]])
        f[i] = L.lub[f[i], min_fi]
    return f


def random_f_monotone_B(L: Lattice, seed=None):
    R = random_state(seed)
    f: Endomorphism = [-1] * L.n
    n_above = L.leq.sum(axis=1)
    p_above = n_above / n_above.sum()
    elems = np.arange(L.n)
    for i in L.irreducibles:
        f[i] = R.choice(elems, 1, p=p_above)[0]
    f[L.bottom] = L.bottom
    for i in L.toposort_bottom_up:
        min_fi = L.lub_of_many(*[f[k] for k in L.children[i]])
        assert all(f[k] != -1 for k in L.children[i])
        if f[i] != -1:
            f[i] = L.lub[f[i], min_fi]
        else:
            f[i] = min_fi
    assert L.f_is_monotone(f)
    return f


def random_f_preserving_lub(L: Lattice, seed: Optional[int] = None,
                            complexity_budget=5):
    f = _random_f_preserving_lub(L, seed=seed)
    if complexity_budget == 0:
        return f
    # Prefer complex functions
    R = random_state(seed)
    seeds = R.randint(0, 2**32, complexity_budget)
    for seed in seeds:
        new_f = _random_f_preserving_lub(L, seed=seed)
        if len(set(f)) < len(set(new_f)):
            f = new_f
    return f


def _random_f_preserving_lub(L: Lattice, seed: Optional[int] = None):
    R = random_state(seed)
    n_above = L.leq.sum(axis=1)
    p_above = n_above / n_above.sum()
    elems = np.arange(L.n)
    f: Endomorphism = [L.bottom] * L.n
    I = np.array(L.irreducibles)
    for i, fi in zip(I, R.choice(elems, len(I), p=p_above)):
        f[i] = fi
    return L.f_preserving_lub(f)



import numpy as np
from .numpy_types import npBoolMatrix, npUInt64Matrix


def floyd_warshall(adj: npBoolMatrix, infinity: int) -> npUInt64Matrix:
    'Compute all pairs shortest distances using Floyd-Warshall algorithm'
    dist: npUInt64Matrix
    dist = adj.astype(np.uint64)  # type:ignore
    dist[~adj] = infinity
    dist[np.diag_indices_from(dist)] = 0
    for k in range(len(dist)):
        np.minimum(dist, dist[:, k, None] + dist[None, k, :], out=dist)
    return dist


import functools
import weakref
from typing import Any, Callable, Optional, Type, TypeVar, Union, cast
from typing_extensions import Protocol
from typing import Callable, TypeVar, Generic
# Cached property both in library and in object

_T = TypeVar('_T')
#import cp93pytools.methodtools


class _Property(Protocol[_T]):
    __get__: Callable[..., _T]


class _NormalMethod(Protocol[_T]):
    __name__: str
    __call__: Callable[..., _T]


def cached_method(maxsize=128, typed=False):
    '''decorator for converting a method into an lru cached method'''

    # https://stackoverflow.com/a/33672499/3671939
    def decorator(func: _NormalMethod[_T]) -> _NormalMethod[_T]:

        @functools.wraps(func)
        def wrapped_func(self, *args, **kwargs):
            # We're storing the wrapped method inside the instance. If we had
            # a strong reference to self the instance would never die.
            weak_self = weakref.ref(self)

            @functools.wraps(func)
            @functools.lru_cache(maxsize=maxsize, typed=typed)
            def _cached_method(*args, **kwargs):
                self = weak_self()
                assert self
                return func(self, *args, **kwargs)

            setattr(self, func.__name__, _cached_method)
            return _cached_method(*args, **kwargs)

        return wrapped_func

    return decorator


class cached_property(property, _Property[_T]):
    '''
    decorator for converting a method into a cached property
    See https://stackoverflow.com/a/4037979/3671939
    This uses a modification:
    1. inherit from property, which disables setattr(instance, name, value)
        as it raises AttributeError: Can't set attribute
    2. use instance.__dict__[name] = value to fix
    '''

    def __init__(self, method: Callable[..., _T]):
        self._method = method

    def __get__(self, instance, _) -> _T:
        name = self._method.__name__
        value = self._method(instance)
        instance.__dict__[name] = value
        return value

    # def delete(self, instance, property_name: str):
    #     del instance.__dict__[property_name]

    @classmethod
    def set_property(cls, instance, property_name: str, value):
        instance.__dict__[property_name] = value

    @classmethod
    def is_computed(cls, instance, property_name: str):
        return property_name in instance.__dict__


class globally_cached_property(cached_property[_T]):
    '''
    decorator for converting a method into a cached property
    See https://stackoverflow.com/a/4037979/3671939
    This uses a modification:
    1. inherit from property, which disables setattr(instance, name, value)
        as it raises AttributeError: Can't set attribute
    2. use instance.__dict__[name] = value to fix
    '''
    cache = {}

    def __get__(self, instance, _) -> _T:
        name = self._method.__name__
        try:
            value = self.cache[(name, instance)]
        except KeyError:
            value = self._method(instance)
            self.cache[(name, instance)] = value
        instance.__dict__[name] = value
        return value

    def set_property(self, instance, property_name: str, value):
        instance.__dict__[property_name] = value
        self.cache[(property_name, instance)] = value



from __future__ import annotations
from typing import Any, Callable, Generic, Iterable, List, Optional, Sequence, Set, Tuple, TypeVar, Union, cast
import numpy as np
import itertools
from .utils.cache_utils import cached_property, cached_method
from .utils.algorithm_tarjan import Tarjan
from .utils.numpy_types import npBoolMatrix, npUInt64Matrix
from .utils.algorithm_floyd_warshall import floyd_warshall
from cp93pytools.methodtools import implemented_at
from . import base_methods as MD
from .base_methods.validation import PosetExceptions

Endomorphism = List[int]
PartialEndomorphism = Union[List[Optional[int]], List[None], Endomorphism]
_T = TypeVar('_T')


class Relation:
    '''
    Class for boolean relation matrices intended mostly for asserting that
    a matrix relation can be used with the fully featured parent classes:
    Poset and Lattice.

    The attribute name leq (less or equal) is missleading for arbitrary relations
    on purpose, because it is mostly used in parent classes.
    '''

    n: int
    leq: npBoolMatrix
    labels: Optional[Sequence[str]]

    MD = MD

    # validation = validation
    # graph = graphs
    # graphviz = graphviz
    # description = description

    def __init__(self, leq: npBoolMatrix, check: bool,
                 labels: Optional[Sequence[str]] = None, **kwargs):
        self.n = len(leq)
        self.leq = leq
        self.labels = labels
        MD.validation.validate_matrix_format(self)
        for k, v in kwargs.items():
            self.__dict__[k] = v
        if check:
            self.validate()
        return

    def validate(self):
        pass  # All binary matrices are valid relations

    @cached_property
    def is_poset(self):
        method = lambda: self.as_poset(check=True)
        return MD.validation.ValidationError.capture(method)

    def as_poset(self, check: bool):
        return cast(Poset, self.copy(cls=Poset, check=check))

    def copy(self, cls=None, check=False):
        cls = cls or self.__class__
        kwargs = self.__dict__.copy()
        kwargs.pop('leq')
        kwargs.pop('labels', None)
        cpy = cls(self.leq, labels=self.labels, check=check, **kwargs)
        return cpy

    def reindex(self, rank: List[int], inverse=False):
        return MD.identity.reindex(self, rank, inverse=inverse)

    @property
    def _labels(self):
        return self.labels or [f'{i}' for i in range(self.n)]

    def relabel(self, labels: Optional[Sequence[str]] = None):
        'copy of self with different labels'
        if labels == self.labels:
            return self
        if labels is not None:
            n = self.n
            m = len(labels)
            assert m == n, f'{m} labels found. Expected {n}'
            non = [l for l in labels if not isinstance(l, str)]
            assert not non, f'non-string label found: {non[0]}'
        Q = self.copy()
        Q.labels = labels
        return Q

    @classmethod
    def total(cls, n: int):
        'total order of n elements'
        G = [[i - 1] if i > 0 else [] for i in range(n)]
        return cls.from_children(G, check=False)  #.as_lattice(check=False)

    '''
    @section
        Interface methods
    '''

    @classmethod
    @implemented_at(MD.interface.from_parents)
    def from_parents(cls, *args, **kwargs):
        ...

    @classmethod
    @implemented_at(MD.interface.from_children)
    def from_children(cls, *args, **kwargs):
        ...

    @classmethod
    @implemented_at(MD.interface.from_down_edges)
    def from_down_edges(cls, *args, **kwargse):
        ...

    @classmethod
    @implemented_at(MD.interface.from_up_edges)
    def from_up_edges(cls, *args, **kwargse):
        ...

    @classmethod
    @implemented_at(MD.interface.from_lambda)
    def from_lambda(cls, *args, **kwargs):
        ...

    # def describe(self):
    #     self.show()
    #     print('Relation matrix:')
    #     print(self.leq.astype(int))
    #     print('Reflexive?', self.is_reflexive)
    #     print('Antisymmetric?', self.is_antisymmetric)
    #     print('Transitive?', self.is_transitive)
    #     return

    # def show(self, labels=None, save=None):
    #     'Display the relation using graphviz. Groups SCCs together'
    #     scc_components, scc_edges = self.scc_reduction()
    #     if labels is None:
    #         labels = [f'{i}' for i in range(self.n)]
    #     n = len(scc_components)
    #     labels = ['-'.join(labels[i] for i in I) for I in scc_components]
    #     return graphviz(n, edges=scc_edges, labels=labels, save=save)
    '''
    @section
        Graph operations
    '''

    def scc_reduction(self):
        n = self.n
        rel = self.leq
        G = [[] for _ in range(n)]
        for i in range(n):
            for j in range(n):
                if rel[i, j] and i != j:
                    G[i].append(j)
        return Tarjan(G).tarjan()

    def transitive_closure(self):
        dist = floyd_warshall(self.leq, infinity=self.n)
        rel = (dist < len(dist))
        rel.flags.writeable = False
        return self.__class__(rel, check=False)

    def __lt_pairs(self):
        for i in range(self.n):
            for j in range(self.n):
                if self.leq[i, j] and i != j:
                    yield i, j

    def f_is_monotone(self, f: Endomorphism):
        for i, j in self.__lt_pairs():
            if not self.leq[f[i], f[j]]:
                return False
        return True


class Poset(Relation):
    '''
    Hashable object that represents an inmutable finite partial order.
    Uses a matrix and hashing is invariant under permutations.

    Run print(Poset.usage) for details and usage examples.

    The main attributes (always present) are:
        n: size of the poset.
            The elements of the poset are range(n)
        leq: read only less-or-equal boolean nxn matrix:
            leq[i,j] is True if and only "i <= j" in the poset order
        
        labels: optional sequence of n strings. Only used for displaying
        _idx: optional permutation of 1..n (not the identity).
    '''

    def validate(self):
        super().validate()
        MD.validation.assert_is_poset(self)

    @cached_property
    def is_lattice(self):
        method = lambda: self.as_lattice(check=True)
        return MD.validation.ValidationError.capture(method)

    def as_lattice(self, check: bool):
        return cast(Lattice, self.copy(cls=Lattice, check=check))

    @cached_property
    def is_distributive(self):
        method = lambda: self.as_distributive(check=True)
        return MD.validation.ValidationError.capture(method)

    def as_distributive(self, check: bool):
        return cast(DistributiveLattice,
                    self.copy(cls=DistributiveLattice, check=check))

    '''
    @section
        Methods for identity
    '''

    def __hash__(self):
        return self.hash

    @cached_property
    def hash(self):
        'hash number for this poset'
        return MD.identity.hasher(sorted(self.hash_elems))

    @cached_property
    def hash_elems(self):
        'hash for each element of the poset w.r.t. the poset order'
        return MD.identity._hash_elems(self, rounds=2, salt=0)

    def __eq__(self, other: Poset):
        'Equality up to isomorphism, i.e. up to reindexing'
        return MD.identity.find_isomorphism(self, other) is not None

    @cached_property
    def hash(self):
        'hash number for this poset'
        return MD.identity.hasher(sorted(self.hash_elems))

    @cached_property
    def canonical(self):
        rank = MD.identity.canonical_rank(self)
        P = self.reindex(rank)
        P.labels = None
        return P

    # __graph = methods_directory.GraphMethods
    # __testing = methods_directory.TestingMethods
    # __generation = methods_directory.GenerationMethods
    # __endomorphism = methods_directory.EndomorphismMethods
    # __initialization = methods_directory.InitializationMethods
    # __binary_operator = methods_directory.BinaryOperatorMethods
    '''
    @section
        Fundamental properties
    '''

    @cached_property
    def child(self) -> npBoolMatrix:
        '''
        nxn boolean matrix: transitive reduction of the poset.
        child[i,j] == True iff j covers i (with no elements inbetween)
        '''
        child = MD.graphs.transitive_reduction(self.leq)
        child.flags.writeable = False
        return child

    @cached_property
    def children(self) -> List[List[int]]:
        ''' top-down adjoint list (j in G[i] iff i covers j)'''
        n = self.n
        child = self.child
        return [[j for j in range(n) if child[j, i]] for i in range(n)]

    @cached_property
    def parents(self) -> List[List[int]]:
        '''bottom-up adjoint list (j in G[i] iff j covers i)'''
        n = self.n
        child = self.child
        return [[j for j in range(n) if child[i, j]] for i in range(n)]

    @cached_property
    def dist(self) -> npUInt64Matrix:
        '''
        Matrix of shortest distance from i upwards to j through parents
        n represents infinity.
        '''
        return MD.graphs.child_to_dist(self.child, assume_poset=True)

    '''
    @section
        Display methods
    '''

    @implemented_at(MD.graphviz.show)
    def show(self):
        ...

    def __repr__(self):
        return self.name

    @cached_property
    def name(self):
        return MD.description.name(self)

    @cached_property
    def upside_down(self):
        geq = self.leq.T
        Q = self.__class__(geq, check=False, labels=self.labels,
                           upside_down=self)
        return Q

    @cached_property
    def toposort_bottom_up(self):
        return MD.graphs.toposort_bottom_up(self)

    @cached_property
    def bottoms(self):
        return MD.graphs.bottoms(self)

    @cached_property
    def tops(self):
        return MD.graphs.tops(self)


class Lattice(Poset):

    def validate(self):
        super().validate()
        MD.validation.assert_is_lattice(self)

    @cached_property
    def lub(self):
        'matrix of i lub j, i.e. i join j'
        n = self.n
        leq = self.leq
        lub_id = {tuple(leq[i, :]): i for i in range(n)}
        lub = np.zeros((n, n), int)
        for i in range(n):
            for j in range(n):
                above = tuple(leq[i, :] & leq[j, :])
                if above not in lub_id:
                    raise PosetExceptions.LUB_Inconsistency(i, j)
                lub[i, j] = lub_id[above]
        lub.flags.writeable = False
        return lub

    @property
    def glb(self):
        return self.upside_down.lub

    @cached_property
    def bottom(self):
        'unique bottom element of the Poset. Throws if not present'
        bottoms = MD.graphs.bottoms(self)
        return MD.validation.expect_unique_bottom(bottoms)

    @cached_property
    def top(self):
        'unique top element of the Poset. Throws if not present'
        tops = MD.graphs.tops(self)
        return MD.validation.expect_unique_top(tops)

    @cached_property
    def irreducibles(self):
        n = self.n
        children = self.children
        return tuple([i for i in range(n) if len(children[i]) == 1])

    @cached_property
    def irreducible_descendants(self):
        '''Irreducibles below x for each x'''
        Rn = range(self.n)
        I = self.irreducibles
        leq = self.leq
        return tuple([tuple([i for i in I if leq[i, x]]) for x in Rn])

    @cached_property
    def irreducible_components(self: Lattice):
        '''
        components of join irreducibles in toposort order and children
        lists for each component
        '''
        n = self.n
        if n <= 1:  # no join irreducibles at all
            return (0, [], [])
        graphs = self.MD.graphs
        irr = self.irreducibles
        sub = graphs.subgraph(self, irr)
        subcomps = graphs.independent_components(sub)
        m = len(subcomps)
        irrcomps = [[irr[j] for j in subcomps[i]] for i in range(m)]
        m_topo, m_children = zip(*(
            MD.graphs._toposort_children(self, irrcomps[i]) for i in range(m)))
        m_topo = cast(Tuple[List[int]], m_topo)
        m_children = cast(Tuple[List[List[int]]], m_children)
        return m, m_topo, m_children

    def lub_of_many(self, *elems: int) -> int:
        if not elems:
            return self.bottom
        lub = self.lub
        acum = elems[0]
        for elem in elems[1:]:
            acum = lub[acum, elem]
        return acum

    def glb_of_many(self, *elems: int) -> int:
        if not elems:
            return self.top
        glb = self.glb
        acum = elems[0]
        for elem in elems[1:]:
            acum = glb[acum, elem]
        return acum

    def f_preserves_lub(self, f: Endomorphism):
        return all(f[self.lub[i, j]] == self.lub[f[i], f[j]]
                   for i in range(self.n)
                   for j in range(self.n))

    def f_lub(self, f: Endomorphism, g: Endomorphism) -> Endomorphism:
        return [self.lub[f[i], g[i]] for i in range(self.n)]

    def f_glb_pointwise(self, f: Endomorphism, g: Endomorphism) -> Endomorphism:
        return [self.glb[f[i], g[i]] for i in range(self.n)]

    def f_glb(self, f: Endomorphism, g: Endomorphism) -> Endomorphism:
        return self.f_preserving_lub(self.f_glb_pointwise(f, g))

    def f_preserving_lub(self, f: Endomorphism,
                         budget: Optional[int] = None) -> Endomorphism:
        '''
        The famous algorithm that fixes iteratively the pairs of
        elements in the lattice that fail to satisfy the LUB axiom
        '''
        lub = self.lub
        leq = self.leq
        it = itertools.count() if budget is None else range(budget)
        for _ in it:
            f_prev = f.copy()
            for i in range(self.n):
                for j in range(self.n):
                    k = lub[i, j]
                    fi_lub_fj = lub[f[i], f[j]]
                    if fi_lub_fj == f[k]:
                        pass
                    elif leq[f[k], fi_lub_fj]:
                        f[k] = fi_lub_fj
                    else:
                        f[i] = lub[f[i], f[k]]
                        f[j] = lub[f[j], f[k]]
            if f == f_prev:
                break
        return f


class DistributiveLattice(Lattice):

    def validate(self):
        super().validate()
        MD.validation.assert_is_distributive(self)